=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2011-11-27 17:50:41 +0000
+++ bzrlib/branch.py	2011-12-05 14:12:23 +0000
@@ -41,6 +41,7 @@
     transport,
     ui,
     urlutils,
+    vf_search,
     )
 from bzrlib.i18n import gettext, ngettext
 """)
@@ -937,7 +938,7 @@
                     tags_to_fetch = set(self.tags.get_reverse_tag_dict())
                 except errors.TagsNotSupported:
                     tags_to_fetch = set()
-                fetch_spec = _mod_graph.NotInOtherForRevs(self.repository,
+                fetch_spec = vf_search.NotInOtherForRevs(self.repository,
                     old_repository, required_ids=[self.last_revision()],
                     if_present_ids=tags_to_fetch, find_ghosts=True).execute()
                 self.repository.fetch(old_repository, fetch_spec=fetch_spec)

=== modified file 'bzrlib/bzrdir.py'
--- bzrlib/bzrdir.py	2011-11-28 10:33:40 +0000
+++ bzrlib/bzrdir.py	2011-12-05 14:12:23 +0000
@@ -46,6 +46,7 @@
     transport as _mod_transport,
     ui,
     urlutils,
+    vf_search,
     win32utils,
     workingtree_3,
     workingtree_4,
@@ -200,7 +201,7 @@
                 if (result_repo.user_url == result.user_url
                     and not require_stacking and
                     revision_id is not None):
-                    fetch_spec = graph.PendingAncestryResult(
+                    fetch_spec = vf_search.PendingAncestryResult(
                         [revision_id], local_repo)
                     result_repo.fetch(local_repo, fetch_spec=fetch_spec)
                 else:

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2011-09-16 15:32:49 +0000
+++ bzrlib/fetch.py	2011-12-05 14:12:23 +0000
@@ -28,9 +28,9 @@
 from bzrlib.lazy_import import lazy_import
 lazy_import(globals(), """
 from bzrlib import (
-    graph as _mod_graph,
     tsort,
     versionedfile,
+    vf_search,
     )
 """)
 from bzrlib import (
@@ -161,13 +161,13 @@
         elif self._last_revision == NULL_REVISION:
             # fetch_spec is None + last_revision is null => empty fetch.
             # explicit limit of no revisions needed
-            return _mod_graph.EmptySearchResult()
+            return vf_search.EmptySearchResult()
         elif self._last_revision is not None:
-            return _mod_graph.NotInOtherForRevs(self.to_repository,
+            return vf_search.NotInOtherForRevs(self.to_repository,
                 self.from_repository, [self._last_revision],
                 find_ghosts=self.find_ghosts).execute()
         else: # self._last_revision is None:
-            return _mod_graph.EverythingNotInOther(self.to_repository,
+            return vf_search.EverythingNotInOther(self.to_repository,
                 self.from_repository,
                 find_ghosts=self.find_ghosts).execute()
 
@@ -389,11 +389,11 @@
                     "limit is only supported with a source branch set")
             # Caller hasn't specified any revisions or source branch
             if self.target_repo_kind == TargetRepoKinds.EMPTY:
-                return _mod_graph.EverythingResult(self.source_repo)
+                return vf_search.EverythingResult(self.source_repo)
             else:
                 # We want everything not already in the target (or target's
                 # fallbacks).
-                return _mod_graph.EverythingNotInOther(
+                return vf_search.EverythingNotInOther(
                     self.target_repo, self.source_repo).execute()
         heads_to_fetch = set(self._explicit_rev_ids)
         if self.source_branch is not None:
@@ -416,7 +416,7 @@
             # heads_to_fetch will almost certainly be present so this doesn't
             # matter much.
             all_heads = heads_to_fetch.union(if_present_fetch)
-            ret = _mod_graph.PendingAncestryResult(all_heads, self.source_repo)
+            ret = vf_search.PendingAncestryResult(all_heads, self.source_repo)
             if self.limit is not None:
                 graph = self.source_repo.get_graph()
                 topo_order = list(graph.iter_topo_order(ret.get_keys()))
@@ -424,6 +424,6 @@
                 ret = self.source_repo.revision_ids_to_search_result(result_set)
             return ret
         else:
-            return _mod_graph.NotInOtherForRevs(self.target_repo, self.source_repo,
+            return vf_search.NotInOtherForRevs(self.target_repo, self.source_repo,
                 required_ids=heads_to_fetch, if_present_ids=if_present_fetch,
                 limit=self.limit).execute()

=== modified file 'bzrlib/graph.py'
--- bzrlib/graph.py	2011-08-26 11:35:50 +0000
+++ bzrlib/graph.py	2011-12-05 15:16:52 +0000
@@ -1346,12 +1346,10 @@
         return ('_BreadthFirstSearcher(iterations=%d, %s,'
                 ' seen=%r)' % (self._iterations, search, list(self.seen)))
 
-    def get_result(self):
-        """Get a SearchResult for the current state of this searcher.
+    def get_state(self):
+        """Get the current state of this searcher.
 
-        :return: A SearchResult for this search so far. The SearchResult is
-            static - the search can be advanced and the search result will not
-            be invalidated or altered.
+        :return: Tuple with started keys, excludes and included keys
         """
         if self._returning == 'next':
             # We have to know the current nodes children to be able to list the
@@ -1368,7 +1366,18 @@
             next_query = self._next_query
         excludes = self._stopped_keys.union(next_query)
         included_keys = self.seen.difference(excludes)
-        return SearchResult(self._started_keys, excludes, len(included_keys),
+        return self._started_keys, excludes, included_keys
+
+    def _get_result(self):
+        """Get a SearchResult for the current state of this searcher.
+
+        :return: A SearchResult for this search so far. The SearchResult is
+            static - the search can be advanced and the search result will not
+            be invalidated or altered.
+        """
+        from bzrlib.vf_search import SearchResult
+        (started_keys, excludes, included_keys) = self.get_state()
+        return SearchResult(started_keys, excludes, len(included_keys),
             included_keys)
 
     def step(self):
@@ -1583,347 +1592,6 @@
             return revs, ghosts
 
 
-class AbstractSearchResult(object):
-    """The result of a search, describing a set of keys.
-    
-    Search results are typically used as the 'fetch_spec' parameter when
-    fetching revisions.
-
-    :seealso: AbstractSearch
-    """
-
-    def get_recipe(self):
-        """Return a recipe that can be used to replay this search.
-
-        The recipe allows reconstruction of the same results at a later date.
-
-        :return: A tuple of `(search_kind_str, *details)`.  The details vary by
-            kind of search result.
-        """
-        raise NotImplementedError(self.get_recipe)
-
-    def get_network_struct(self):
-        """Return a tuple that can be transmitted via the HPSS protocol."""
-        raise NotImplementedError(self.get_network_struct)
-
-    def get_keys(self):
-        """Return the keys found in this search.
-
-        :return: A set of keys.
-        """
-        raise NotImplementedError(self.get_keys)
-
-    def is_empty(self):
-        """Return false if the search lists 1 or more revisions."""
-        raise NotImplementedError(self.is_empty)
-
-    def refine(self, seen, referenced):
-        """Create a new search by refining this search.
-
-        :param seen: Revisions that have been satisfied.
-        :param referenced: Revision references observed while satisfying some
-            of this search.
-        :return: A search result.
-        """
-        raise NotImplementedError(self.refine)
-
-
-class AbstractSearch(object):
-    """A search that can be executed, producing a search result.
-
-    :seealso: AbstractSearchResult
-    """
-
-    def execute(self):
-        """Construct a network-ready search result from this search description.
-
-        This may take some time to search repositories, etc.
-
-        :return: A search result (an object that implements
-            AbstractSearchResult's API).
-        """
-        raise NotImplementedError(self.execute)
-
-
-class SearchResult(AbstractSearchResult):
-    """The result of a breadth first search.
-
-    A SearchResult provides the ability to reconstruct the search or access a
-    set of the keys the search found.
-    """
-
-    def __init__(self, start_keys, exclude_keys, key_count, keys):
-        """Create a SearchResult.
-
-        :param start_keys: The keys the search started at.
-        :param exclude_keys: The keys the search excludes.
-        :param key_count: The total number of keys (from start to but not
-            including exclude).
-        :param keys: The keys the search found. Note that in future we may get
-            a SearchResult from a smart server, in which case the keys list is
-            not necessarily immediately available.
-        """
-        self._recipe = ('search', start_keys, exclude_keys, key_count)
-        self._keys = frozenset(keys)
-
-    def __repr__(self):
-        kind, start_keys, exclude_keys, key_count = self._recipe
-        if len(start_keys) > 5:
-            start_keys_repr = repr(list(start_keys)[:5])[:-1] + ', ...]'
-        else:
-            start_keys_repr = repr(start_keys)
-        if len(exclude_keys) > 5:
-            exclude_keys_repr = repr(list(exclude_keys)[:5])[:-1] + ', ...]'
-        else:
-            exclude_keys_repr = repr(exclude_keys)
-        return '<%s %s:(%s, %s, %d)>' % (self.__class__.__name__,
-            kind, start_keys_repr, exclude_keys_repr, key_count)
-
-    def get_recipe(self):
-        """Return a recipe that can be used to replay this search.
-
-        The recipe allows reconstruction of the same results at a later date
-        without knowing all the found keys. The essential elements are a list
-        of keys to start and to stop at. In order to give reproducible
-        results when ghosts are encountered by a search they are automatically
-        added to the exclude list (or else ghost filling may alter the
-        results).
-
-        :return: A tuple ('search', start_keys_set, exclude_keys_set,
-            revision_count). To recreate the results of this search, create a
-            breadth first searcher on the same graph starting at start_keys.
-            Then call next() (or next_with_ghosts()) repeatedly, and on every
-            result, call stop_searching_any on any keys from the exclude_keys
-            set. The revision_count value acts as a trivial cross-check - the
-            found revisions of the new search should have as many elements as
-            revision_count. If it does not, then additional revisions have been
-            ghosted since the search was executed the first time and the second
-            time.
-        """
-        return self._recipe
-
-    def get_network_struct(self):
-        start_keys = ' '.join(self._recipe[1])
-        stop_keys = ' '.join(self._recipe[2])
-        count = str(self._recipe[3])
-        return (self._recipe[0], '\n'.join((start_keys, stop_keys, count)))
-
-    def get_keys(self):
-        """Return the keys found in this search.
-
-        :return: A set of keys.
-        """
-        return self._keys
-
-    def is_empty(self):
-        """Return false if the search lists 1 or more revisions."""
-        return self._recipe[3] == 0
-
-    def refine(self, seen, referenced):
-        """Create a new search by refining this search.
-
-        :param seen: Revisions that have been satisfied.
-        :param referenced: Revision references observed while satisfying some
-            of this search.
-        """
-        start = self._recipe[1]
-        exclude = self._recipe[2]
-        count = self._recipe[3]
-        keys = self.get_keys()
-        # New heads = referenced + old heads - seen things - exclude
-        pending_refs = set(referenced)
-        pending_refs.update(start)
-        pending_refs.difference_update(seen)
-        pending_refs.difference_update(exclude)
-        # New exclude = old exclude + satisfied heads
-        seen_heads = start.intersection(seen)
-        exclude.update(seen_heads)
-        # keys gets seen removed
-        keys = keys - seen
-        # length is reduced by len(seen)
-        count -= len(seen)
-        return SearchResult(pending_refs, exclude, count, keys)
-
-
-class PendingAncestryResult(AbstractSearchResult):
-    """A search result that will reconstruct the ancestry for some graph heads.
-
-    Unlike SearchResult, this doesn't hold the complete search result in
-    memory, it just holds a description of how to generate it.
-    """
-
-    def __init__(self, heads, repo):
-        """Constructor.
-
-        :param heads: an iterable of graph heads.
-        :param repo: a repository to use to generate the ancestry for the given
-            heads.
-        """
-        self.heads = frozenset(heads)
-        self.repo = repo
-
-    def __repr__(self):
-        if len(self.heads) > 5:
-            heads_repr = repr(list(self.heads)[:5])[:-1]
-            heads_repr += ', <%d more>...]' % (len(self.heads) - 5,)
-        else:
-            heads_repr = repr(self.heads)
-        return '<%s heads:%s repo:%r>' % (
-            self.__class__.__name__, heads_repr, self.repo)
-
-    def get_recipe(self):
-        """Return a recipe that can be used to replay this search.
-
-        The recipe allows reconstruction of the same results at a later date.
-
-        :seealso SearchResult.get_recipe:
-
-        :return: A tuple ('proxy-search', start_keys_set, set(), -1)
-            To recreate this result, create a PendingAncestryResult with the
-            start_keys_set.
-        """
-        return ('proxy-search', self.heads, set(), -1)
-
-    def get_network_struct(self):
-        parts = ['ancestry-of']
-        parts.extend(self.heads)
-        return parts
-
-    def get_keys(self):
-        """See SearchResult.get_keys.
-
-        Returns all the keys for the ancestry of the heads, excluding
-        NULL_REVISION.
-        """
-        return self._get_keys(self.repo.get_graph())
-
-    def _get_keys(self, graph):
-        NULL_REVISION = revision.NULL_REVISION
-        keys = [key for (key, parents) in graph.iter_ancestry(self.heads)
-                if key != NULL_REVISION and parents is not None]
-        return keys
-
-    def is_empty(self):
-        """Return false if the search lists 1 or more revisions."""
-        if revision.NULL_REVISION in self.heads:
-            return len(self.heads) == 1
-        else:
-            return len(self.heads) == 0
-
-    def refine(self, seen, referenced):
-        """Create a new search by refining this search.
-
-        :param seen: Revisions that have been satisfied.
-        :param referenced: Revision references observed while satisfying some
-            of this search.
-        """
-        referenced = self.heads.union(referenced)
-        return PendingAncestryResult(referenced - seen, self.repo)
-
-
-class EmptySearchResult(AbstractSearchResult):
-    """An empty search result."""
-
-    def is_empty(self):
-        return True
-    
-
-class EverythingResult(AbstractSearchResult):
-    """A search result that simply requests everything in the repository."""
-
-    def __init__(self, repo):
-        self._repo = repo
-
-    def __repr__(self):
-        return '%s(%r)' % (self.__class__.__name__, self._repo)
-
-    def get_recipe(self):
-        raise NotImplementedError(self.get_recipe)
-
-    def get_network_struct(self):
-        return ('everything',)
-
-    def get_keys(self):
-        if 'evil' in debug.debug_flags:
-            from bzrlib import remote
-            if isinstance(self._repo, remote.RemoteRepository):
-                # warn developers (not users) not to do this
-                trace.mutter_callsite(
-                    2, "EverythingResult(RemoteRepository).get_keys() is slow.")
-        return self._repo.all_revision_ids()
-
-    def is_empty(self):
-        # It's ok for this to wrongly return False: the worst that can happen
-        # is that RemoteStreamSource will initiate a get_stream on an empty
-        # repository.  And almost all repositories are non-empty.
-        return False
-
-    def refine(self, seen, referenced):
-        heads = set(self._repo.all_revision_ids())
-        heads.difference_update(seen)
-        heads.update(referenced)
-        return PendingAncestryResult(heads, self._repo)
-
-
-class EverythingNotInOther(AbstractSearch):
-    """Find all revisions in that are in one repo but not the other."""
-
-    def __init__(self, to_repo, from_repo, find_ghosts=False):
-        self.to_repo = to_repo
-        self.from_repo = from_repo
-        self.find_ghosts = find_ghosts
-
-    def execute(self):
-        return self.to_repo.search_missing_revision_ids(
-            self.from_repo, find_ghosts=self.find_ghosts)
-
-
-class NotInOtherForRevs(AbstractSearch):
-    """Find all revisions missing in one repo for a some specific heads."""
-
-    def __init__(self, to_repo, from_repo, required_ids, if_present_ids=None,
-            find_ghosts=False, limit=None):
-        """Constructor.
-
-        :param required_ids: revision IDs of heads that must be found, or else
-            the search will fail with NoSuchRevision.  All revisions in their
-            ancestry not already in the other repository will be included in
-            the search result.
-        :param if_present_ids: revision IDs of heads that may be absent in the
-            source repository.  If present, then their ancestry not already
-            found in other will be included in the search result.
-        :param limit: maximum number of revisions to fetch
-        """
-        self.to_repo = to_repo
-        self.from_repo = from_repo
-        self.find_ghosts = find_ghosts
-        self.required_ids = required_ids
-        self.if_present_ids = if_present_ids
-        self.limit = limit
-
-    def __repr__(self):
-        if len(self.required_ids) > 5:
-            reqd_revs_repr = repr(list(self.required_ids)[:5])[:-1] + ', ...]'
-        else:
-            reqd_revs_repr = repr(self.required_ids)
-        if self.if_present_ids and len(self.if_present_ids) > 5:
-            ifp_revs_repr = repr(list(self.if_present_ids)[:5])[:-1] + ', ...]'
-        else:
-            ifp_revs_repr = repr(self.if_present_ids)
-
-        return ("<%s from:%r to:%r find_ghosts:%r req'd:%r if-present:%r"
-                "limit:%r>") % (
-                self.__class__.__name__, self.from_repo, self.to_repo,
-                self.find_ghosts, reqd_revs_repr, ifp_revs_repr,
-                self.limit)
-
-    def execute(self):
-        return self.to_repo.search_missing_revision_ids(
-            self.from_repo, revision_ids=self.required_ids,
-            if_present_ids=self.if_present_ids, find_ghosts=self.find_ghosts,
-            limit=self.limit)
-
-
 def invert_parent_map(parent_map):
     """Given a map from child => parents, create a map of parent=>children"""
     child_map = {}
@@ -1939,145 +1607,6 @@
     return child_map
 
 
-def _find_possible_heads(parent_map, tip_keys, depth):
-    """Walk backwards (towards children) through the parent_map.
-
-    This finds 'heads' that will hopefully succinctly describe our search
-    graph.
-    """
-    child_map = invert_parent_map(parent_map)
-    heads = set()
-    current_roots = tip_keys
-    walked = set(current_roots)
-    while current_roots and depth > 0:
-        depth -= 1
-        children = set()
-        children_update = children.update
-        for p in current_roots:
-            # Is it better to pre- or post- filter the children?
-            try:
-                children_update(child_map[p])
-            except KeyError:
-                heads.add(p)
-        # If we've seen a key before, we don't want to walk it again. Note that
-        # 'children' stays relatively small while 'walked' grows large. So
-        # don't use 'difference_update' here which has to walk all of 'walked'.
-        # '.difference' is smart enough to walk only children and compare it to
-        # walked.
-        children = children.difference(walked)
-        walked.update(children)
-        current_roots = children
-    if current_roots:
-        # We walked to the end of depth, so these are the new tips.
-        heads.update(current_roots)
-    return heads
-
-
-def _run_search(parent_map, heads, exclude_keys):
-    """Given a parent map, run a _BreadthFirstSearcher on it.
-
-    Start at heads, walk until you hit exclude_keys. As a further improvement,
-    watch for any heads that you encounter while walking, which means they were
-    not heads of the search.
-
-    This is mostly used to generate a succinct recipe for how to walk through
-    most of parent_map.
-
-    :return: (_BreadthFirstSearcher, set(heads_encountered_by_walking))
-    """
-    g = Graph(DictParentsProvider(parent_map))
-    s = g._make_breadth_first_searcher(heads)
-    found_heads = set()
-    while True:
-        try:
-            next_revs = s.next()
-        except StopIteration:
-            break
-        for parents in s._current_parents.itervalues():
-            f_heads = heads.intersection(parents)
-            if f_heads:
-                found_heads.update(f_heads)
-        stop_keys = exclude_keys.intersection(next_revs)
-        if stop_keys:
-            s.stop_searching_any(stop_keys)
-    for parents in s._current_parents.itervalues():
-        f_heads = heads.intersection(parents)
-        if f_heads:
-            found_heads.update(f_heads)
-    return s, found_heads
-
-
-def limited_search_result_from_parent_map(parent_map, missing_keys, tip_keys,
-                                          depth):
-    """Transform a parent_map that is searching 'tip_keys' into an
-    approximate SearchResult.
-
-    We should be able to generate a SearchResult from a given set of starting
-    keys, that covers a subset of parent_map that has the last step pointing at
-    tip_keys. This is to handle the case that really-long-searches shouldn't be
-    started from scratch on each get_parent_map request, but we *do* want to
-    filter out some of the keys that we've already seen, so we don't get
-    information that we already know about on every request.
-
-    The server will validate the search (that starting at start_keys and
-    stopping at stop_keys yields the exact key_count), so we have to be careful
-    to give an exact recipe.
-
-    Basic algorithm is:
-        1) Invert parent_map to get child_map (todo: have it cached and pass it
-           in)
-        2) Starting at tip_keys, walk towards children for 'depth' steps.
-        3) At that point, we have the 'start' keys.
-        4) Start walking parent_map from 'start' keys, counting how many keys
-           are seen, and generating stop_keys for anything that would walk
-           outside of the parent_map.
-
-    :param parent_map: A map from {child_id: (parent_ids,)}
-    :param missing_keys: parent_ids that we know are unavailable
-    :param tip_keys: the revision_ids that we are searching
-    :param depth: How far back to walk.
-    """
-    if not parent_map:
-        # No search to send, because we haven't done any searching yet.
-        return [], [], 0
-    heads = _find_possible_heads(parent_map, tip_keys, depth)
-    s, found_heads = _run_search(parent_map, heads, set(tip_keys))
-    _, start_keys, exclude_keys, key_count = s.get_result().get_recipe()
-    if found_heads:
-        # Anything in found_heads are redundant start_keys, we hit them while
-        # walking, so we can exclude them from the start list.
-        start_keys = set(start_keys).difference(found_heads)
-    return start_keys, exclude_keys, key_count
-
-
-def search_result_from_parent_map(parent_map, missing_keys):
-    """Transform a parent_map into SearchResult information."""
-    if not parent_map:
-        # parent_map is empty or None, simple search result
-        return [], [], 0
-    # start_set is all the keys in the cache
-    start_set = set(parent_map)
-    # result set is all the references to keys in the cache
-    result_parents = set()
-    for parents in parent_map.itervalues():
-        result_parents.update(parents)
-    stop_keys = result_parents.difference(start_set)
-    # We don't need to send ghosts back to the server as a position to
-    # stop either.
-    stop_keys.difference_update(missing_keys)
-    key_count = len(parent_map)
-    if (revision.NULL_REVISION in result_parents
-        and revision.NULL_REVISION in missing_keys):
-        # If we pruned NULL_REVISION from the stop_keys because it's also
-        # in our cache of "missing" keys we need to increment our key count
-        # by 1, because the reconsitituted SearchResult on the server will
-        # still consider NULL_REVISION to be an included key.
-        key_count += 1
-    included_keys = start_set.intersection(result_parents)
-    start_set.difference_update(included_keys)
-    return start_set, stop_keys, key_count
-
-
 def collapse_linear_regions(parent_map):
     """Collapse regions of the graph that are 'linear'.
 

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2011-11-28 15:59:39 +0000
+++ bzrlib/remote.py	2011-12-05 14:12:23 +0000
@@ -38,6 +38,7 @@
     testament as _mod_testament,
     urlutils,
     vf_repository,
+    vf_search,
     )
 from bzrlib.branch import BranchReferenceFormat, BranchWriteLockResult
 from bzrlib.decorators import needs_read_lock, needs_write_lock, only_raises
@@ -1841,7 +1842,7 @@
         included_keys = result_set.intersection(result_parents)
         start_keys = result_set.difference(included_keys)
         exclude_keys = result_parents.difference(result_set)
-        result = graph.SearchResult(start_keys, exclude_keys,
+        result = vf_search.SearchResult(start_keys, exclude_keys,
             len(result_set), result_set)
         return result
 
@@ -1895,6 +1896,10 @@
         # the InterRepository base class, which raises an
         # IncompatibleRepositories when asked to fetch.
         inter = _mod_repository.InterRepository.get(source, self)
+        if (fetch_spec is not None and
+            not getattr(inter, "supports_fetch_spec", False)):
+            raise errors.UnsupportedOperation(
+                "fetch_spec not supported for %r" % inter)
         return inter.fetch(revision_id=revision_id,
             find_ghosts=find_ghosts, fetch_spec=fetch_spec)
 
@@ -2062,11 +2067,11 @@
             parents_map = {}
         if _DEFAULT_SEARCH_DEPTH <= 0:
             (start_set, stop_keys,
-             key_count) = graph.search_result_from_parent_map(
+             key_count) = vf_search.search_result_from_parent_map(
                 parents_map, self._unstacked_provider.missing_keys)
         else:
             (start_set, stop_keys,
-             key_count) = graph.limited_search_result_from_parent_map(
+             key_count) = vf_search.limited_search_result_from_parent_map(
                 parents_map, self._unstacked_provider.missing_keys,
                 keys, depth=_DEFAULT_SEARCH_DEPTH)
         recipe = ('manual', start_set, stop_keys, key_count)
@@ -2684,7 +2689,7 @@
             except errors.UnknownSmartMethod:
                 medium._remember_remote_is_before(version)
             except errors.UnknownErrorFromSmartServer, e:
-                if isinstance(search, graph.EverythingResult):
+                if isinstance(search, vf_search.EverythingResult):
                     error_verb = e.error_from_smart_server.error_verb
                     if error_verb == 'BadSearch':
                         # Pre-2.4 servers don't support this sort of search.

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2011-11-19 17:39:55 +0000
+++ bzrlib/repository.py	2011-12-05 14:12:23 +0000
@@ -675,12 +675,10 @@
     def _resume_write_group(self, tokens):
         raise errors.UnsuspendableWriteGroup(self)
 
-    def fetch(self, source, revision_id=None, find_ghosts=False,
-            fetch_spec=None):
+    def fetch(self, source, revision_id=None, find_ghosts=False):
         """Fetch the content required to construct revision_id from source.
 
-        If revision_id is None and fetch_spec is None, then all content is
-        copied.
+        If revision_id is None, then all content is copied.
 
         fetch() may not be used when the repository is in a write group -
         either finish the current write group before using fetch, or use
@@ -692,14 +690,7 @@
         :param revision_id: If specified, all the content needed for this
             revision ID will be copied to the target.  Fetch will determine for
             itself which content needs to be copied.
-        :param fetch_spec: If specified, a SearchResult or
-            PendingAncestryResult that describes which revisions to copy.  This
-            allows copying multiple heads at once.  Mutually exclusive with
-            revision_id.
         """
-        if fetch_spec is not None and revision_id is not None:
-            raise AssertionError(
-                "fetch_spec and revision_id are mutually exclusive.")
         if self.is_in_write_group():
             raise errors.InternalBzrError(
                 "May not fetch while in a write group.")
@@ -707,7 +698,6 @@
         # TODO: lift out to somewhere common with RemoteRepository
         # <https://bugs.launchpad.net/bzr/+bug/401646>
         if (self.has_same_location(source)
-            and fetch_spec is None
             and self._has_same_fallbacks(source)):
             # check that last_revision is in 'from' and then return a
             # no-operation.
@@ -716,8 +706,7 @@
                 self.get_revision(revision_id)
             return 0, []
         inter = InterRepository.get(source, self)
-        return inter.fetch(revision_id=revision_id,
-            find_ghosts=find_ghosts, fetch_spec=fetch_spec)
+        return inter.fetch(revision_id=revision_id, find_ghosts=find_ghosts)
 
     def create_bundle(self, target, base, fileobj, format=None):
         return serializer.write_bundle(self, target, base, fileobj, format)
@@ -1158,19 +1147,6 @@
                 [parents_provider, other_repository._make_parents_provider()])
         return graph.Graph(parents_provider)
 
-    def revision_ids_to_search_result(self, result_set):
-        """Convert a set of revision ids to a graph SearchResult."""
-        result_parents = set()
-        for parents in self.get_graph().get_parent_map(
-            result_set).itervalues():
-            result_parents.update(parents)
-        included_keys = result_set.intersection(result_parents)
-        start_keys = result_set.difference(included_keys)
-        exclude_keys = result_parents.difference(result_set)
-        result = graph.SearchResult(start_keys, exclude_keys,
-            len(result_set), result_set)
-        return result
-
     @needs_write_lock
     def set_make_working_trees(self, new_value):
         """Set the policy flag for making working trees when creating branches.
@@ -1716,8 +1692,7 @@
         self.target.fetch(self.source, revision_id=revision_id)
 
     @needs_write_lock
-    def fetch(self, revision_id=None, find_ghosts=False,
-            fetch_spec=None):
+    def fetch(self, revision_id=None, find_ghosts=False):
         """Fetch the content required to construct revision_id.
 
         The content is copied from self.source to self.target.

=== modified file 'bzrlib/smart/repository.py'
--- bzrlib/smart/repository.py	2011-11-28 15:59:39 +0000
+++ bzrlib/smart/repository.py	2011-12-05 15:16:52 +0000
@@ -28,11 +28,11 @@
     bencode,
     errors,
     estimate_compressed_size,
-    graph,
     osutils,
     pack,
     trace,
     ui,
+    vf_search,
     )
 from bzrlib.bzrdir import BzrDir
 from bzrlib.smart.request import (
@@ -85,11 +85,11 @@
             they expected and get it from elsewhere.
         """
         if search_bytes == 'everything':
-            return graph.EverythingResult(repository), None
+            return vf_search.EverythingResult(repository), None
         lines = search_bytes.split('\n')
         if lines[0] == 'ancestry-of':
             heads = lines[1:]
-            search_result = graph.PendingAncestryResult(heads, repository)
+            search_result = vf_search.PendingAncestryResult(heads, repository)
             return search_result, None
         elif lines[0] == 'search':
             return self.recreate_search_from_recipe(repository, lines[1:],
@@ -119,15 +119,16 @@
                 except StopIteration:
                     break
                 search.stop_searching_any(exclude_keys.intersection(next_revs))
-            search_result = search.get_result()
-            if (not discard_excess and
-                search_result.get_recipe()[3] != revision_count):
+            (started_keys, excludes, included_keys) = search.get_state()
+            if (not discard_excess and len(included_keys) != revision_count):
                 # we got back a different amount of data than expected, this
                 # gets reported as NoSuchRevision, because less revisions
                 # indicates missing revisions, and more should never happen as
                 # the excludes list considers ghosts and ensures that ghost
                 # filling races are not a problem.
                 return (None, FailedSmartServerResponse(('NoSuchRevision',)))
+            search_result = vf_search.SearchResult(started_keys, excludes,
+                len(included_keys), included_keys)
             return (search_result, None)
         finally:
             repository.unlock()

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2011-12-05 14:21:55 +0000
+++ bzrlib/tests/__init__.py	2011-12-06 16:05:01 +0000
@@ -4093,6 +4093,7 @@
         'bzrlib.tests.test_version',
         'bzrlib.tests.test_version_info',
         'bzrlib.tests.test_versionedfile',
+        'bzrlib.tests.test_vf_search',
         'bzrlib.tests.test_weave',
         'bzrlib.tests.test_whitebox',
         'bzrlib.tests.test_win32utils',

=== modified file 'bzrlib/tests/per_interrepository/test_fetch.py'
--- bzrlib/tests/per_interrepository/test_fetch.py	2011-05-08 13:07:11 +0000
+++ bzrlib/tests/per_interrepository/test_fetch.py	2011-12-05 14:12:23 +0000
@@ -27,7 +27,7 @@
 from bzrlib.errors import (
     NoSuchRevision,
     )
-from bzrlib.graph import (
+from bzrlib.vf_search import (
     SearchResult,
     )
 from bzrlib.revision import (

=== modified file 'bzrlib/tests/per_repository_reference/test_fetch.py'
--- bzrlib/tests/per_repository_reference/test_fetch.py	2010-12-06 04:59:38 +0000
+++ bzrlib/tests/per_repository_reference/test_fetch.py	2011-12-05 14:12:23 +0000
@@ -17,11 +17,7 @@
 
 from bzrlib import (
     branch,
-    errors,
-    graph,
-    )
-from bzrlib.smart import (
-    server,
+    vf_search,
     )
 from bzrlib.tests.per_repository import TestCaseWithRepository
 
@@ -176,7 +172,7 @@
         self.addCleanup(target.lock_write().unlock)
         target.fetch(
             repo_missing_fallbacks,
-            fetch_spec=graph.EverythingResult(repo_missing_fallbacks))
+            fetch_spec=vf_search.EverythingResult(repo_missing_fallbacks))
         self.assertEqual(repo_missing_fallbacks.revisions.keys(),
             target.revisions.keys())
         self.assertEqual(repo_missing_fallbacks.inventories.keys(),

=== modified file 'bzrlib/tests/per_repository_vf/test_fetch.py'
--- bzrlib/tests/per_repository_vf/test_fetch.py	2011-05-02 22:44:40 +0000
+++ bzrlib/tests/per_repository_vf/test_fetch.py	2011-12-05 14:12:23 +0000
@@ -17,7 +17,7 @@
 """Tests for fetch between repositories of the same type."""
 
 from bzrlib import (
-    graph,
+    vf_search,
     )
 from bzrlib.tests.per_repository_vf import (
     TestCaseWithRepository,
@@ -49,7 +49,7 @@
         self.addCleanup(b.unlock)
         repo = b.repository
         source = repo._get_source(repo._format)
-        search = graph.PendingAncestryResult(['tip'], repo)
+        search = vf_search.PendingAncestryResult(['tip'], repo)
         stream = source.get_stream(search)
         for substream_type, substream in stream:
             for record in substream:

=== modified file 'bzrlib/tests/test_graph.py'
--- bzrlib/tests/test_graph.py	2011-08-26 11:35:50 +0000
+++ bzrlib/tests/test_graph.py	2011-12-05 15:16:52 +0000
@@ -1029,9 +1029,8 @@
                 search.start_searching(starts)
             if stops is not None:
                 search.stop_searching_any(stops)
-            result = search.get_result()
-            self.assertEqual(recipe, result.get_recipe())
-            self.assertEqual(set(included_keys), result.get_keys())
+            state = search.get_state()
+            self.assertEqual(set(included_keys), state[2])
             self.assertEqual(seen, search.seen)
 
     def test_breadth_first_get_result_excludes_current_pending(self):
@@ -1042,10 +1041,9 @@
             })
         search = graph._make_breadth_first_searcher(['head'])
         # At the start, nothing has been seen, to its all excluded:
-        result = search.get_result()
-        self.assertEqual(('search', set(['head']), set(['head']), 0),
-            result.get_recipe())
-        self.assertEqual(set(), result.get_keys())
+        state = search.get_state()
+        self.assertEqual((set(['head']), set(['head']), set()),
+            state)
         self.assertEqual(set(), search.seen)
         # using next:
         expected = [
@@ -1074,10 +1072,9 @@
         # Starting with nothing and adding a search works:
         search.start_searching(['head'])
         # head has been seen:
-        result = search.get_result()
-        self.assertEqual(('search', set(['head']), set(['child']), 1),
-            result.get_recipe())
-        self.assertEqual(set(['head']), result.get_keys())
+        state = search.get_state()
+        self.assertEqual((set(['head']), set(['child']), set(['head'])),
+            state)
         self.assertEqual(set(['head']), search.seen)
         # using next:
         expected = [
@@ -1234,19 +1231,21 @@
         self.assertSeenAndResult(expected, search, search.next)
         self.assertRaises(StopIteration, search.next)
         self.assertEqual(set(['head', 'ghost', NULL_REVISION]), search.seen)
-        result = search.get_result()
-        self.assertEqual(('search', set(['ghost', 'head']), set(['ghost']), 2),
-            result.get_recipe())
-        self.assertEqual(set(['head', NULL_REVISION]), result.get_keys())
+        state = search.get_state()
+        self.assertEqual(
+            (set(['ghost', 'head']), set(['ghost']),
+                set(['head', NULL_REVISION])),
+            state)
         # using next_with_ghosts:
         search = graph._make_breadth_first_searcher(['head'])
         self.assertSeenAndResult(expected, search, search.next_with_ghosts)
         self.assertRaises(StopIteration, search.next)
         self.assertEqual(set(['head', 'ghost', NULL_REVISION]), search.seen)
-        result = search.get_result()
-        self.assertEqual(('search', set(['ghost', 'head']), set(['ghost']), 2),
-            result.get_recipe())
-        self.assertEqual(set(['head', NULL_REVISION]), result.get_keys())
+        state = search.get_state()
+        self.assertEqual(
+            (set(['ghost', 'head']), set(['ghost']),
+                set(['head', NULL_REVISION])),
+            state)
 
 
 class TestFindUniqueAncestors(TestGraphBase):
@@ -1677,167 +1676,6 @@
                  for n in graph_thunk.merge_sort('C')])
 
 
-class TestPendingAncestryResultGetKeys(TestCaseWithMemoryTransport):
-    """Tests for bzrlib.graph.PendingAncestryResult."""
-
-    def test_get_keys(self):
-        builder = self.make_branch_builder('b')
-        builder.start_series()
-        builder.build_snapshot('rev-1', None, [
-            ('add', ('', 'root-id', 'directory', ''))])
-        builder.build_snapshot('rev-2', ['rev-1'], [])
-        builder.finish_series()
-        repo = builder.get_branch().repository
-        repo.lock_read()
-        self.addCleanup(repo.unlock)
-        result = _mod_graph.PendingAncestryResult(['rev-2'], repo)
-        self.assertEqual(set(['rev-1', 'rev-2']), set(result.get_keys()))
-
-    def test_get_keys_excludes_ghosts(self):
-        builder = self.make_branch_builder('b')
-        builder.start_series()
-        builder.build_snapshot('rev-1', None, [
-            ('add', ('', 'root-id', 'directory', ''))])
-        builder.build_snapshot('rev-2', ['rev-1', 'ghost'], [])
-        builder.finish_series()
-        repo = builder.get_branch().repository
-        repo.lock_read()
-        self.addCleanup(repo.unlock)
-        result = _mod_graph.PendingAncestryResult(['rev-2'], repo)
-        self.assertEqual(sorted(['rev-1', 'rev-2']), sorted(result.get_keys()))
-
-    def test_get_keys_excludes_null(self):
-        # Make a 'graph' with an iter_ancestry that returns NULL_REVISION
-        # somewhere other than the last element, which can happen in real
-        # ancestries.
-        class StubGraph(object):
-            def iter_ancestry(self, keys):
-                return [(NULL_REVISION, ()), ('foo', (NULL_REVISION,))]
-        result = _mod_graph.PendingAncestryResult(['rev-3'], None)
-        result_keys = result._get_keys(StubGraph())
-        # Only the non-null keys from the ancestry appear.
-        self.assertEqual(set(['foo']), set(result_keys))
-
-
-class TestPendingAncestryResultRefine(TestGraphBase):
-
-    def test_refine(self):
-        # Used when pulling from a stacked repository, so test some revisions
-        # being satisfied from the stacking branch.
-        g = self.make_graph(
-            {"tip":["mid"], "mid":["base"], "tag":["base"],
-             "base":[NULL_REVISION], NULL_REVISION:[]})
-        result = _mod_graph.PendingAncestryResult(['tip', 'tag'], None)
-        result = result.refine(set(['tip']), set(['mid']))
-        self.assertEqual(set(['mid', 'tag']), result.heads)
-        result = result.refine(set(['mid', 'tag', 'base']),
-            set([NULL_REVISION]))
-        self.assertEqual(set([NULL_REVISION]), result.heads)
-        self.assertTrue(result.is_empty())
-
-
-class TestSearchResultRefine(TestGraphBase):
-
-    def test_refine(self):
-        # Used when pulling from a stacked repository, so test some revisions
-        # being satisfied from the stacking branch.
-        g = self.make_graph(
-            {"tip":["mid"], "mid":["base"], "tag":["base"],
-             "base":[NULL_REVISION], NULL_REVISION:[]})
-        result = _mod_graph.SearchResult(set(['tip', 'tag']),
-            set([NULL_REVISION]), 4, set(['tip', 'mid', 'tag', 'base']))
-        result = result.refine(set(['tip']), set(['mid']))
-        recipe = result.get_recipe()
-        # We should be starting from tag (original head) and mid (seen ref)
-        self.assertEqual(set(['mid', 'tag']), recipe[1])
-        # We should be stopping at NULL (original stop) and tip (seen head)
-        self.assertEqual(set([NULL_REVISION, 'tip']), recipe[2])
-        self.assertEqual(3, recipe[3])
-        result = result.refine(set(['mid', 'tag', 'base']),
-            set([NULL_REVISION]))
-        recipe = result.get_recipe()
-        # We should be starting from nothing (NULL was known as a cut point)
-        self.assertEqual(set([]), recipe[1])
-        # We should be stopping at NULL (original stop) and tip (seen head) and
-        # tag (seen head) and mid(seen mid-point head). We could come back and
-        # define this as not including mid, for minimal results, but it is
-        # still 'correct' to include mid, and simpler/easier.
-        self.assertEqual(set([NULL_REVISION, 'tip', 'tag', 'mid']), recipe[2])
-        self.assertEqual(0, recipe[3])
-        self.assertTrue(result.is_empty())
-
-
-class TestSearchResultFromParentMap(TestGraphBase):
-
-    def assertSearchResult(self, start_keys, stop_keys, key_count, parent_map,
-                           missing_keys=()):
-        (start, stop, count) = _mod_graph.search_result_from_parent_map(
-            parent_map, missing_keys)
-        self.assertEqual((sorted(start_keys), sorted(stop_keys), key_count),
-                         (sorted(start), sorted(stop), count))
-
-    def test_no_parents(self):
-        self.assertSearchResult([], [], 0, {})
-        self.assertSearchResult([], [], 0, None)
-
-    def test_ancestry_1(self):
-        self.assertSearchResult(['rev4'], [NULL_REVISION], len(ancestry_1),
-                                ancestry_1)
-
-    def test_ancestry_2(self):
-        self.assertSearchResult(['rev1b', 'rev4a'], [NULL_REVISION],
-                                len(ancestry_2), ancestry_2)
-        self.assertSearchResult(['rev1b', 'rev4a'], [],
-                                len(ancestry_2)+1, ancestry_2,
-                                missing_keys=[NULL_REVISION])
-
-    def test_partial_search(self):
-        parent_map = dict((k,extended_history_shortcut[k])
-                          for k in ['e', 'f'])
-        self.assertSearchResult(['e', 'f'], ['d', 'a'], 2,
-                                parent_map)
-        parent_map.update((k,extended_history_shortcut[k])
-                          for k in ['d', 'a'])
-        self.assertSearchResult(['e', 'f'], ['c', NULL_REVISION], 4,
-                                parent_map)
-        parent_map['c'] = extended_history_shortcut['c']
-        self.assertSearchResult(['e', 'f'], ['b'], 6,
-                                parent_map, missing_keys=[NULL_REVISION])
-        parent_map['b'] = extended_history_shortcut['b']
-        self.assertSearchResult(['e', 'f'], [], 7,
-                                parent_map, missing_keys=[NULL_REVISION])
-
-
-class TestLimitedSearchResultFromParentMap(TestGraphBase):
-
-    def assertSearchResult(self, start_keys, stop_keys, key_count, parent_map,
-                           missing_keys, tip_keys, depth):
-        (start, stop, count) = _mod_graph.limited_search_result_from_parent_map(
-            parent_map, missing_keys, tip_keys, depth)
-        self.assertEqual((sorted(start_keys), sorted(stop_keys), key_count),
-                         (sorted(start), sorted(stop), count))
-
-    def test_empty_ancestry(self):
-        self.assertSearchResult([], [], 0, {}, (), ['tip-rev-id'], 10)
-
-    def test_ancestry_1(self):
-        self.assertSearchResult(['rev4'], ['rev1'], 4,
-                                ancestry_1, (), ['rev1'], 10)
-        self.assertSearchResult(['rev2a', 'rev2b'], ['rev1'], 2,
-                                ancestry_1, (), ['rev1'], 1)
-
-
-    def test_multiple_heads(self):
-        self.assertSearchResult(['e', 'f'], ['a'], 5,
-                                extended_history_shortcut, (), ['a'], 10)
-        # Note that even though we only take 1 step back, we find 'f', which
-        # means the described search will still find d and c.
-        self.assertSearchResult(['f'], ['a'], 4,
-                                extended_history_shortcut, (), ['a'], 1)
-        self.assertSearchResult(['f'], ['a'], 4,
-                                extended_history_shortcut, (), ['a'], 2)
-
-
 class TestStackedParentsProvider(tests.TestCase):
 
     def setUp(self):

=== modified file 'bzrlib/tests/test_reconfigure.py'
--- bzrlib/tests/test_reconfigure.py	2011-06-10 08:11:44 +0000
+++ bzrlib/tests/test_reconfigure.py	2011-12-05 17:31:43 +0000
@@ -21,6 +21,7 @@
     reconfigure,
     repository,
     tests,
+    vf_repository,
     workingtree,
     )
 
@@ -258,12 +259,12 @@
     def test_branch_to_lightweight_checkout_failure(self):
         parent, child, reconfiguration = \
             self.prepare_branch_to_lightweight_checkout()
-        old_Repository_fetch = repository.Repository.fetch
-        repository.Repository.fetch = None
+        old_Repository_fetch = vf_repository.VersionedFileRepository.fetch
+        vf_repository.VersionedFileRepository.fetch = None
         try:
             self.assertRaises(TypeError, reconfiguration.apply)
         finally:
-            repository.Repository.fetch = old_Repository_fetch
+            vf_repository.VersionedFileRepository.fetch = old_Repository_fetch
         child = _mod_branch.Branch.open('child')
         self.assertContainsRe(child.base, 'child/$')
 

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2011-11-28 17:30:38 +0000
+++ bzrlib/tests/test_remote.py	2011-12-05 14:12:23 +0000
@@ -42,6 +42,7 @@
     transport,
     treebuilder,
     versionedfile,
+    vf_search,
     )
 from bzrlib.branch import Branch
 from bzrlib.bzrdir import (
@@ -3891,7 +3892,7 @@
         revs = [r for (r,ps) in graph.iter_ancestry([tip])
                 if r != NULL_REVISION]
         revs.reverse()
-        search = _mod_graph.PendingAncestryResult([tip], stacked.repository)
+        search = vf_search.PendingAncestryResult([tip], stacked.repository)
         self.reset_smart_call_log()
         stream = source.get_stream(search)
         # We trust that if a revision is in the stream the rest of the new
@@ -4003,7 +4004,7 @@
         self.hpss_calls = []
         local.repository.fetch(
             remote_branch.repository,
-            fetch_spec=_mod_graph.EverythingResult(remote_branch.repository))
+            fetch_spec=vf_search.EverythingResult(remote_branch.repository))
         self.assertEqual(['Repository.get_stream_1.19'], self.hpss_calls)
 
     def override_verb(self, verb_name, verb):
@@ -4043,7 +4044,7 @@
         self.hpss_calls = []
         local.repository.fetch(
             remote_branch.repository,
-            fetch_spec=_mod_graph.EverythingResult(remote_branch.repository))
+            fetch_spec=vf_search.EverythingResult(remote_branch.repository))
         # make sure the overridden verb was used
         self.assertLength(1, verb_log)
         # more than one HPSS call is needed, but because it's a VFS callback

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2011-09-16 20:42:31 +0000
+++ bzrlib/tests/test_repository.py	2011-12-05 14:12:23 +0000
@@ -31,10 +31,10 @@
     )
 from bzrlib import (
     btree_index,
-    graph,
     symbol_versioning,
     tests,
     transport,
+    vf_search,
     )
 from bzrlib.btree_index import BTreeBuilder, BTreeGraphIndex
 from bzrlib.index import GraphIndex
@@ -706,7 +706,7 @@
 
         # On a regular pass, getting the inventories and chk pages for rev-2
         # would only get the newly created chk pages
-        search = graph.SearchResult(set(['rev-2']), set(['rev-1']), 1,
+        search = vf_search.SearchResult(set(['rev-2']), set(['rev-1']), 1,
                                     set(['rev-2']))
         simple_chk_records = []
         for vf_name, substream in source.get_stream(search):

=== added file 'bzrlib/tests/test_vf_search.py'
--- bzrlib/tests/test_vf_search.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_vf_search.py	2011-12-05 14:12:23 +0000
@@ -0,0 +1,239 @@
+# Copyright (C) 2007-2011 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+
+from bzrlib import (
+    graph as _mod_graph,
+    tests,
+    vf_search,
+    )
+from bzrlib.revision import NULL_REVISION
+from bzrlib.tests.test_graph import TestGraphBase
+
+# Ancestry 1:
+#
+#  NULL_REVISION
+#       |
+#     rev1
+#      /\
+#  rev2a rev2b
+#     |    |
+#   rev3  /
+#     |  /
+#   rev4
+ancestry_1 = {'rev1': [NULL_REVISION], 'rev2a': ['rev1'], 'rev2b': ['rev1'],
+              'rev3': ['rev2a'], 'rev4': ['rev3', 'rev2b']}
+
+# Ancestry 2:
+#
+#  NULL_REVISION
+#    /    \
+# rev1a  rev1b
+#   |
+# rev2a
+#   |
+# rev3a
+#   |
+# rev4a
+ancestry_2 = {'rev1a': [NULL_REVISION], 'rev2a': ['rev1a'],
+              'rev1b': [NULL_REVISION], 'rev3a': ['rev2a'], 'rev4a': ['rev3a']}
+
+
+# Extended history shortcut
+#  NULL_REVISION
+#       |
+#       a
+#       |\
+#       b |
+#       | |
+#       c |
+#       | |
+#       d |
+#       |\|
+#       e f
+extended_history_shortcut = {'a': [NULL_REVISION],
+                             'b': ['a'],
+                             'c': ['b'],
+                             'd': ['c'],
+                             'e': ['d'],
+                             'f': ['a', 'd'],
+                            }
+
+
+class TestSearchResultRefine(tests.TestCase):
+
+    def make_graph(self, ancestors):
+        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))
+
+    def test_refine(self):
+        # Used when pulling from a stacked repository, so test some revisions
+        # being satisfied from the stacking branch.
+        g = self.make_graph(
+            {"tip":["mid"], "mid":["base"], "tag":["base"],
+             "base":[NULL_REVISION], NULL_REVISION:[]})
+        result = vf_search.SearchResult(set(['tip', 'tag']),
+            set([NULL_REVISION]), 4, set(['tip', 'mid', 'tag', 'base']))
+        result = result.refine(set(['tip']), set(['mid']))
+        recipe = result.get_recipe()
+        # We should be starting from tag (original head) and mid (seen ref)
+        self.assertEqual(set(['mid', 'tag']), recipe[1])
+        # We should be stopping at NULL (original stop) and tip (seen head)
+        self.assertEqual(set([NULL_REVISION, 'tip']), recipe[2])
+        self.assertEqual(3, recipe[3])
+        result = result.refine(set(['mid', 'tag', 'base']),
+            set([NULL_REVISION]))
+        recipe = result.get_recipe()
+        # We should be starting from nothing (NULL was known as a cut point)
+        self.assertEqual(set([]), recipe[1])
+        # We should be stopping at NULL (original stop) and tip (seen head) and
+        # tag (seen head) and mid(seen mid-point head). We could come back and
+        # define this as not including mid, for minimal results, but it is
+        # still 'correct' to include mid, and simpler/easier.
+        self.assertEqual(set([NULL_REVISION, 'tip', 'tag', 'mid']), recipe[2])
+        self.assertEqual(0, recipe[3])
+        self.assertTrue(result.is_empty())
+
+
+class TestSearchResultFromParentMap(TestGraphBase):
+
+    def assertSearchResult(self, start_keys, stop_keys, key_count, parent_map,
+                           missing_keys=()):
+        (start, stop, count) = vf_search.search_result_from_parent_map(
+            parent_map, missing_keys)
+        self.assertEqual((sorted(start_keys), sorted(stop_keys), key_count),
+                         (sorted(start), sorted(stop), count))
+
+    def test_no_parents(self):
+        self.assertSearchResult([], [], 0, {})
+        self.assertSearchResult([], [], 0, None)
+
+    def test_ancestry_1(self):
+        self.assertSearchResult(['rev4'], [NULL_REVISION], len(ancestry_1),
+                                ancestry_1)
+
+    def test_ancestry_2(self):
+        self.assertSearchResult(['rev1b', 'rev4a'], [NULL_REVISION],
+                                len(ancestry_2), ancestry_2)
+        self.assertSearchResult(['rev1b', 'rev4a'], [],
+                                len(ancestry_2)+1, ancestry_2,
+                                missing_keys=[NULL_REVISION])
+
+    def test_partial_search(self):
+        parent_map = dict((k,extended_history_shortcut[k])
+                          for k in ['e', 'f'])
+        self.assertSearchResult(['e', 'f'], ['d', 'a'], 2,
+                                parent_map)
+        parent_map.update((k,extended_history_shortcut[k])
+                          for k in ['d', 'a'])
+        self.assertSearchResult(['e', 'f'], ['c', NULL_REVISION], 4,
+                                parent_map)
+        parent_map['c'] = extended_history_shortcut['c']
+        self.assertSearchResult(['e', 'f'], ['b'], 6,
+                                parent_map, missing_keys=[NULL_REVISION])
+        parent_map['b'] = extended_history_shortcut['b']
+        self.assertSearchResult(['e', 'f'], [], 7,
+                                parent_map, missing_keys=[NULL_REVISION])
+
+
+class TestLimitedSearchResultFromParentMap(TestGraphBase):
+
+    def assertSearchResult(self, start_keys, stop_keys, key_count, parent_map,
+                           missing_keys, tip_keys, depth):
+        (start, stop, count) = vf_search.limited_search_result_from_parent_map(
+            parent_map, missing_keys, tip_keys, depth)
+        self.assertEqual((sorted(start_keys), sorted(stop_keys), key_count),
+                         (sorted(start), sorted(stop), count))
+
+    def test_empty_ancestry(self):
+        self.assertSearchResult([], [], 0, {}, (), ['tip-rev-id'], 10)
+
+    def test_ancestry_1(self):
+        self.assertSearchResult(['rev4'], ['rev1'], 4,
+                                ancestry_1, (), ['rev1'], 10)
+        self.assertSearchResult(['rev2a', 'rev2b'], ['rev1'], 2,
+                                ancestry_1, (), ['rev1'], 1)
+
+
+    def test_multiple_heads(self):
+        self.assertSearchResult(['e', 'f'], ['a'], 5,
+                                extended_history_shortcut, (), ['a'], 10)
+        # Note that even though we only take 1 step back, we find 'f', which
+        # means the described search will still find d and c.
+        self.assertSearchResult(['f'], ['a'], 4,
+                                extended_history_shortcut, (), ['a'], 1)
+        self.assertSearchResult(['f'], ['a'], 4,
+                                extended_history_shortcut, (), ['a'], 2)
+
+
+class TestPendingAncestryResultRefine(tests.TestCase):
+
+    def make_graph(self, ancestors):
+        return _mod_graph.Graph(_mod_graph.DictParentsProvider(ancestors))
+
+    def test_refine(self):
+        # Used when pulling from a stacked repository, so test some revisions
+        # being satisfied from the stacking branch.
+        g = self.make_graph(
+            {"tip":["mid"], "mid":["base"], "tag":["base"],
+             "base":[NULL_REVISION], NULL_REVISION:[]})
+        result = vf_search.PendingAncestryResult(['tip', 'tag'], None)
+        result = result.refine(set(['tip']), set(['mid']))
+        self.assertEqual(set(['mid', 'tag']), result.heads)
+        result = result.refine(set(['mid', 'tag', 'base']),
+            set([NULL_REVISION]))
+        self.assertEqual(set([NULL_REVISION]), result.heads)
+        self.assertTrue(result.is_empty())
+
+
+class TestPendingAncestryResultGetKeys(tests.TestCaseWithMemoryTransport):
+    """Tests for bzrlib.graph.PendingAncestryResult."""
+
+    def test_get_keys(self):
+        builder = self.make_branch_builder('b')
+        builder.start_series()
+        builder.build_snapshot('rev-1', None, [
+            ('add', ('', 'root-id', 'directory', ''))])
+        builder.build_snapshot('rev-2', ['rev-1'], [])
+        builder.finish_series()
+        repo = builder.get_branch().repository
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
+        result = vf_search.PendingAncestryResult(['rev-2'], repo)
+        self.assertEqual(set(['rev-1', 'rev-2']), set(result.get_keys()))
+
+    def test_get_keys_excludes_ghosts(self):
+        builder = self.make_branch_builder('b')
+        builder.start_series()
+        builder.build_snapshot('rev-1', None, [
+            ('add', ('', 'root-id', 'directory', ''))])
+        builder.build_snapshot('rev-2', ['rev-1', 'ghost'], [])
+        builder.finish_series()
+        repo = builder.get_branch().repository
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
+        result = vf_search.PendingAncestryResult(['rev-2'], repo)
+        self.assertEqual(sorted(['rev-1', 'rev-2']), sorted(result.get_keys()))
+
+    def test_get_keys_excludes_null(self):
+        # Make a 'graph' with an iter_ancestry that returns NULL_REVISION
+        # somewhere other than the last element, which can happen in real
+        # ancestries.
+        class StubGraph(object):
+            def iter_ancestry(self, keys):
+                return [(NULL_REVISION, ()), ('foo', (NULL_REVISION,))]
+        result = vf_search.PendingAncestryResult(['rev-3'], None)
+        result_keys = result._get_keys(StubGraph())
+        # Only the non-null keys from the ancestry appear.
+        self.assertEqual(set(['foo']), set(result_keys))

=== modified file 'bzrlib/vf_repository.py'
--- bzrlib/vf_repository.py	2011-12-03 03:30:10 +0000
+++ bzrlib/vf_repository.py	2011-12-06 16:05:01 +0000
@@ -38,6 +38,7 @@
     tsort,
     ui,
     versionedfile,
+    vf_search,
     )
 
 from bzrlib.recordcounter import RecordCounter
@@ -1220,6 +1221,54 @@
         # rather copying them?
         self._safe_to_return_from_cache = False
 
+    def fetch(self, source, revision_id=None, find_ghosts=False,
+            fetch_spec=None):
+        """Fetch the content required to construct revision_id from source.
+
+        If revision_id is None and fetch_spec is None, then all content is
+        copied.
+
+        fetch() may not be used when the repository is in a write group -
+        either finish the current write group before using fetch, or use
+        fetch before starting the write group.
+
+        :param find_ghosts: Find and copy revisions in the source that are
+            ghosts in the target (and not reachable directly by walking out to
+            the first-present revision in target from revision_id).
+        :param revision_id: If specified, all the content needed for this
+            revision ID will be copied to the target.  Fetch will determine for
+            itself which content needs to be copied.
+        :param fetch_spec: If specified, a SearchResult or
+            PendingAncestryResult that describes which revisions to copy.  This
+            allows copying multiple heads at once.  Mutually exclusive with
+            revision_id.
+        """
+        if fetch_spec is not None and revision_id is not None:
+            raise AssertionError(
+                "fetch_spec and revision_id are mutually exclusive.")
+        if self.is_in_write_group():
+            raise errors.InternalBzrError(
+                "May not fetch while in a write group.")
+        # fast path same-url fetch operations
+        # TODO: lift out to somewhere common with RemoteRepository
+        # <https://bugs.launchpad.net/bzr/+bug/401646>
+        if (self.has_same_location(source)
+            and fetch_spec is None
+            and self._has_same_fallbacks(source)):
+            # check that last_revision is in 'from' and then return a
+            # no-operation.
+            if (revision_id is not None and
+                not _mod_revision.is_null(revision_id)):
+                self.get_revision(revision_id)
+            return 0, []
+        inter = InterRepository.get(source, self)
+        if (fetch_spec is not None and
+            not getattr(inter, "supports_fetch_spec", False)):
+            raise errors.UnsupportedOperation(
+                "fetch_spec not supported for %r" % inter)
+        return inter.fetch(revision_id=revision_id,
+            find_ghosts=find_ghosts, fetch_spec=fetch_spec)
+
     @needs_read_lock
     def gather_stats(self, revid=None, committers=None):
         """See Repository.gather_stats()."""
@@ -1854,6 +1903,19 @@
         """Return the graph walker for text revisions."""
         return graph.Graph(self.texts)
 
+    def revision_ids_to_search_result(self, result_set):
+        """Convert a set of revision ids to a graph SearchResult."""
+        result_parents = set()
+        for parents in self.get_graph().get_parent_map(
+            result_set).itervalues():
+            result_parents.update(parents)
+        included_keys = result_set.intersection(result_parents)
+        start_keys = result_set.difference(included_keys)
+        exclude_keys = result_parents.difference(result_set)
+        result = vf_search.SearchResult(start_keys, exclude_keys,
+            len(result_set), result_set)
+        return result
+
     def _get_versioned_file_checker(self, text_key_references=None,
         ancestors=None):
         """Return an object suitable for checking versioned files.
@@ -2495,6 +2557,8 @@
 
     _walk_to_common_revisions_batch_size = 50
 
+    supports_fetch_spec = True
+
     @needs_write_lock
     def fetch(self, revision_id=None, find_ghosts=False,
             fetch_spec=None):
@@ -2576,7 +2640,9 @@
                 searcher.stop_searching_any(stop_revs)
             if searcher_exhausted:
                 break
-        return searcher.get_result()
+        (started_keys, excludes, included_keys) = searcher.get_state()
+        return vf_search.SearchResult(started_keys, excludes,
+            len(included_keys), included_keys)
 
     @needs_read_lock
     def search_missing_revision_ids(self,

=== added file 'bzrlib/vf_search.py'
--- bzrlib/vf_search.py	1970-01-01 00:00:00 +0000
+++ bzrlib/vf_search.py	2011-12-05 15:16:52 +0000
@@ -0,0 +1,509 @@
+# Copyright (C) 2007-2011 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+
+"""Searching in versioned file repositories."""
+
+from bzrlib import (
+    debug,
+    revision,
+    trace,
+    )
+
+from bzrlib.graph import (
+    DictParentsProvider,
+    Graph,
+    invert_parent_map,
+    )
+
+
+class AbstractSearchResult(object):
+    """The result of a search, describing a set of keys.
+    
+    Search results are typically used as the 'fetch_spec' parameter when
+    fetching revisions.
+
+    :seealso: AbstractSearch
+    """
+
+    def get_recipe(self):
+        """Return a recipe that can be used to replay this search.
+
+        The recipe allows reconstruction of the same results at a later date.
+
+        :return: A tuple of `(search_kind_str, *details)`.  The details vary by
+            kind of search result.
+        """
+        raise NotImplementedError(self.get_recipe)
+
+    def get_network_struct(self):
+        """Return a tuple that can be transmitted via the HPSS protocol."""
+        raise NotImplementedError(self.get_network_struct)
+
+    def get_keys(self):
+        """Return the keys found in this search.
+
+        :return: A set of keys.
+        """
+        raise NotImplementedError(self.get_keys)
+
+    def is_empty(self):
+        """Return false if the search lists 1 or more revisions."""
+        raise NotImplementedError(self.is_empty)
+
+    def refine(self, seen, referenced):
+        """Create a new search by refining this search.
+
+        :param seen: Revisions that have been satisfied.
+        :param referenced: Revision references observed while satisfying some
+            of this search.
+        :return: A search result.
+        """
+        raise NotImplementedError(self.refine)
+
+
+class AbstractSearch(object):
+    """A search that can be executed, producing a search result.
+
+    :seealso: AbstractSearchResult
+    """
+
+    def execute(self):
+        """Construct a network-ready search result from this search description.
+
+        This may take some time to search repositories, etc.
+
+        :return: A search result (an object that implements
+            AbstractSearchResult's API).
+        """
+        raise NotImplementedError(self.execute)
+
+
+class SearchResult(AbstractSearchResult):
+    """The result of a breadth first search.
+
+    A SearchResult provides the ability to reconstruct the search or access a
+    set of the keys the search found.
+    """
+
+    def __init__(self, start_keys, exclude_keys, key_count, keys):
+        """Create a SearchResult.
+
+        :param start_keys: The keys the search started at.
+        :param exclude_keys: The keys the search excludes.
+        :param key_count: The total number of keys (from start to but not
+            including exclude).
+        :param keys: The keys the search found. Note that in future we may get
+            a SearchResult from a smart server, in which case the keys list is
+            not necessarily immediately available.
+        """
+        self._recipe = ('search', start_keys, exclude_keys, key_count)
+        self._keys = frozenset(keys)
+
+    def __repr__(self):
+        kind, start_keys, exclude_keys, key_count = self._recipe
+        if len(start_keys) > 5:
+            start_keys_repr = repr(list(start_keys)[:5])[:-1] + ', ...]'
+        else:
+            start_keys_repr = repr(start_keys)
+        if len(exclude_keys) > 5:
+            exclude_keys_repr = repr(list(exclude_keys)[:5])[:-1] + ', ...]'
+        else:
+            exclude_keys_repr = repr(exclude_keys)
+        return '<%s %s:(%s, %s, %d)>' % (self.__class__.__name__,
+            kind, start_keys_repr, exclude_keys_repr, key_count)
+
+    def get_recipe(self):
+        """Return a recipe that can be used to replay this search.
+
+        The recipe allows reconstruction of the same results at a later date
+        without knowing all the found keys. The essential elements are a list
+        of keys to start and to stop at. In order to give reproducible
+        results when ghosts are encountered by a search they are automatically
+        added to the exclude list (or else ghost filling may alter the
+        results).
+
+        :return: A tuple ('search', start_keys_set, exclude_keys_set,
+            revision_count). To recreate the results of this search, create a
+            breadth first searcher on the same graph starting at start_keys.
+            Then call next() (or next_with_ghosts()) repeatedly, and on every
+            result, call stop_searching_any on any keys from the exclude_keys
+            set. The revision_count value acts as a trivial cross-check - the
+            found revisions of the new search should have as many elements as
+            revision_count. If it does not, then additional revisions have been
+            ghosted since the search was executed the first time and the second
+            time.
+        """
+        return self._recipe
+
+    def get_network_struct(self):
+        start_keys = ' '.join(self._recipe[1])
+        stop_keys = ' '.join(self._recipe[2])
+        count = str(self._recipe[3])
+        return (self._recipe[0], '\n'.join((start_keys, stop_keys, count)))
+
+    def get_keys(self):
+        """Return the keys found in this search.
+
+        :return: A set of keys.
+        """
+        return self._keys
+
+    def is_empty(self):
+        """Return false if the search lists 1 or more revisions."""
+        return self._recipe[3] == 0
+
+    def refine(self, seen, referenced):
+        """Create a new search by refining this search.
+
+        :param seen: Revisions that have been satisfied.
+        :param referenced: Revision references observed while satisfying some
+            of this search.
+        """
+        start = self._recipe[1]
+        exclude = self._recipe[2]
+        count = self._recipe[3]
+        keys = self.get_keys()
+        # New heads = referenced + old heads - seen things - exclude
+        pending_refs = set(referenced)
+        pending_refs.update(start)
+        pending_refs.difference_update(seen)
+        pending_refs.difference_update(exclude)
+        # New exclude = old exclude + satisfied heads
+        seen_heads = start.intersection(seen)
+        exclude.update(seen_heads)
+        # keys gets seen removed
+        keys = keys - seen
+        # length is reduced by len(seen)
+        count -= len(seen)
+        return SearchResult(pending_refs, exclude, count, keys)
+
+
+class PendingAncestryResult(AbstractSearchResult):
+    """A search result that will reconstruct the ancestry for some graph heads.
+
+    Unlike SearchResult, this doesn't hold the complete search result in
+    memory, it just holds a description of how to generate it.
+    """
+
+    def __init__(self, heads, repo):
+        """Constructor.
+
+        :param heads: an iterable of graph heads.
+        :param repo: a repository to use to generate the ancestry for the given
+            heads.
+        """
+        self.heads = frozenset(heads)
+        self.repo = repo
+
+    def __repr__(self):
+        if len(self.heads) > 5:
+            heads_repr = repr(list(self.heads)[:5])[:-1]
+            heads_repr += ', <%d more>...]' % (len(self.heads) - 5,)
+        else:
+            heads_repr = repr(self.heads)
+        return '<%s heads:%s repo:%r>' % (
+            self.__class__.__name__, heads_repr, self.repo)
+
+    def get_recipe(self):
+        """Return a recipe that can be used to replay this search.
+
+        The recipe allows reconstruction of the same results at a later date.
+
+        :seealso SearchResult.get_recipe:
+
+        :return: A tuple ('proxy-search', start_keys_set, set(), -1)
+            To recreate this result, create a PendingAncestryResult with the
+            start_keys_set.
+        """
+        return ('proxy-search', self.heads, set(), -1)
+
+    def get_network_struct(self):
+        parts = ['ancestry-of']
+        parts.extend(self.heads)
+        return parts
+
+    def get_keys(self):
+        """See SearchResult.get_keys.
+
+        Returns all the keys for the ancestry of the heads, excluding
+        NULL_REVISION.
+        """
+        return self._get_keys(self.repo.get_graph())
+
+    def _get_keys(self, graph):
+        NULL_REVISION = revision.NULL_REVISION
+        keys = [key for (key, parents) in graph.iter_ancestry(self.heads)
+                if key != NULL_REVISION and parents is not None]
+        return keys
+
+    def is_empty(self):
+        """Return false if the search lists 1 or more revisions."""
+        if revision.NULL_REVISION in self.heads:
+            return len(self.heads) == 1
+        else:
+            return len(self.heads) == 0
+
+    def refine(self, seen, referenced):
+        """Create a new search by refining this search.
+
+        :param seen: Revisions that have been satisfied.
+        :param referenced: Revision references observed while satisfying some
+            of this search.
+        """
+        referenced = self.heads.union(referenced)
+        return PendingAncestryResult(referenced - seen, self.repo)
+
+
+class EmptySearchResult(AbstractSearchResult):
+    """An empty search result."""
+
+    def is_empty(self):
+        return True
+
+
+class EverythingResult(AbstractSearchResult):
+    """A search result that simply requests everything in the repository."""
+
+    def __init__(self, repo):
+        self._repo = repo
+
+    def __repr__(self):
+        return '%s(%r)' % (self.__class__.__name__, self._repo)
+
+    def get_recipe(self):
+        raise NotImplementedError(self.get_recipe)
+
+    def get_network_struct(self):
+        return ('everything',)
+
+    def get_keys(self):
+        if 'evil' in debug.debug_flags:
+            from bzrlib import remote
+            if isinstance(self._repo, remote.RemoteRepository):
+                # warn developers (not users) not to do this
+                trace.mutter_callsite(
+                    2, "EverythingResult(RemoteRepository).get_keys() is slow.")
+        return self._repo.all_revision_ids()
+
+    def is_empty(self):
+        # It's ok for this to wrongly return False: the worst that can happen
+        # is that RemoteStreamSource will initiate a get_stream on an empty
+        # repository.  And almost all repositories are non-empty.
+        return False
+
+    def refine(self, seen, referenced):
+        heads = set(self._repo.all_revision_ids())
+        heads.difference_update(seen)
+        heads.update(referenced)
+        return PendingAncestryResult(heads, self._repo)
+
+
+class EverythingNotInOther(AbstractSearch):
+    """Find all revisions in that are in one repo but not the other."""
+
+    def __init__(self, to_repo, from_repo, find_ghosts=False):
+        self.to_repo = to_repo
+        self.from_repo = from_repo
+        self.find_ghosts = find_ghosts
+
+    def execute(self):
+        return self.to_repo.search_missing_revision_ids(
+            self.from_repo, find_ghosts=self.find_ghosts)
+
+
+class NotInOtherForRevs(AbstractSearch):
+    """Find all revisions missing in one repo for a some specific heads."""
+
+    def __init__(self, to_repo, from_repo, required_ids, if_present_ids=None,
+            find_ghosts=False, limit=None):
+        """Constructor.
+
+        :param required_ids: revision IDs of heads that must be found, or else
+            the search will fail with NoSuchRevision.  All revisions in their
+            ancestry not already in the other repository will be included in
+            the search result.
+        :param if_present_ids: revision IDs of heads that may be absent in the
+            source repository.  If present, then their ancestry not already
+            found in other will be included in the search result.
+        :param limit: maximum number of revisions to fetch
+        """
+        self.to_repo = to_repo
+        self.from_repo = from_repo
+        self.find_ghosts = find_ghosts
+        self.required_ids = required_ids
+        self.if_present_ids = if_present_ids
+        self.limit = limit
+
+    def __repr__(self):
+        if len(self.required_ids) > 5:
+            reqd_revs_repr = repr(list(self.required_ids)[:5])[:-1] + ', ...]'
+        else:
+            reqd_revs_repr = repr(self.required_ids)
+        if self.if_present_ids and len(self.if_present_ids) > 5:
+            ifp_revs_repr = repr(list(self.if_present_ids)[:5])[:-1] + ', ...]'
+        else:
+            ifp_revs_repr = repr(self.if_present_ids)
+
+        return ("<%s from:%r to:%r find_ghosts:%r req'd:%r if-present:%r"
+                "limit:%r>") % (
+                self.__class__.__name__, self.from_repo, self.to_repo,
+                self.find_ghosts, reqd_revs_repr, ifp_revs_repr,
+                self.limit)
+
+    def execute(self):
+        return self.to_repo.search_missing_revision_ids(
+            self.from_repo, revision_ids=self.required_ids,
+            if_present_ids=self.if_present_ids, find_ghosts=self.find_ghosts,
+            limit=self.limit)
+
+
+def search_result_from_parent_map(parent_map, missing_keys):
+    """Transform a parent_map into SearchResult information."""
+    if not parent_map:
+        # parent_map is empty or None, simple search result
+        return [], [], 0
+    # start_set is all the keys in the cache
+    start_set = set(parent_map)
+    # result set is all the references to keys in the cache
+    result_parents = set()
+    for parents in parent_map.itervalues():
+        result_parents.update(parents)
+    stop_keys = result_parents.difference(start_set)
+    # We don't need to send ghosts back to the server as a position to
+    # stop either.
+    stop_keys.difference_update(missing_keys)
+    key_count = len(parent_map)
+    if (revision.NULL_REVISION in result_parents
+        and revision.NULL_REVISION in missing_keys):
+        # If we pruned NULL_REVISION from the stop_keys because it's also
+        # in our cache of "missing" keys we need to increment our key count
+        # by 1, because the reconsitituted SearchResult on the server will
+        # still consider NULL_REVISION to be an included key.
+        key_count += 1
+    included_keys = start_set.intersection(result_parents)
+    start_set.difference_update(included_keys)
+    return start_set, stop_keys, key_count
+
+
+def _run_search(parent_map, heads, exclude_keys):
+    """Given a parent map, run a _BreadthFirstSearcher on it.
+
+    Start at heads, walk until you hit exclude_keys. As a further improvement,
+    watch for any heads that you encounter while walking, which means they were
+    not heads of the search.
+
+    This is mostly used to generate a succinct recipe for how to walk through
+    most of parent_map.
+
+    :return: (_BreadthFirstSearcher, set(heads_encountered_by_walking))
+    """
+    g = Graph(DictParentsProvider(parent_map))
+    s = g._make_breadth_first_searcher(heads)
+    found_heads = set()
+    while True:
+        try:
+            next_revs = s.next()
+        except StopIteration:
+            break
+        for parents in s._current_parents.itervalues():
+            f_heads = heads.intersection(parents)
+            if f_heads:
+                found_heads.update(f_heads)
+        stop_keys = exclude_keys.intersection(next_revs)
+        if stop_keys:
+            s.stop_searching_any(stop_keys)
+    for parents in s._current_parents.itervalues():
+        f_heads = heads.intersection(parents)
+        if f_heads:
+            found_heads.update(f_heads)
+    return s, found_heads
+
+
+def _find_possible_heads(parent_map, tip_keys, depth):
+    """Walk backwards (towards children) through the parent_map.
+
+    This finds 'heads' that will hopefully succinctly describe our search
+    graph.
+    """
+    child_map = invert_parent_map(parent_map)
+    heads = set()
+    current_roots = tip_keys
+    walked = set(current_roots)
+    while current_roots and depth > 0:
+        depth -= 1
+        children = set()
+        children_update = children.update
+        for p in current_roots:
+            # Is it better to pre- or post- filter the children?
+            try:
+                children_update(child_map[p])
+            except KeyError:
+                heads.add(p)
+        # If we've seen a key before, we don't want to walk it again. Note that
+        # 'children' stays relatively small while 'walked' grows large. So
+        # don't use 'difference_update' here which has to walk all of 'walked'.
+        # '.difference' is smart enough to walk only children and compare it to
+        # walked.
+        children = children.difference(walked)
+        walked.update(children)
+        current_roots = children
+    if current_roots:
+        # We walked to the end of depth, so these are the new tips.
+        heads.update(current_roots)
+    return heads
+
+
+def limited_search_result_from_parent_map(parent_map, missing_keys, tip_keys,
+                                          depth):
+    """Transform a parent_map that is searching 'tip_keys' into an
+    approximate SearchResult.
+
+    We should be able to generate a SearchResult from a given set of starting
+    keys, that covers a subset of parent_map that has the last step pointing at
+    tip_keys. This is to handle the case that really-long-searches shouldn't be
+    started from scratch on each get_parent_map request, but we *do* want to
+    filter out some of the keys that we've already seen, so we don't get
+    information that we already know about on every request.
+
+    The server will validate the search (that starting at start_keys and
+    stopping at stop_keys yields the exact key_count), so we have to be careful
+    to give an exact recipe.
+
+    Basic algorithm is:
+        1) Invert parent_map to get child_map (todo: have it cached and pass it
+           in)
+        2) Starting at tip_keys, walk towards children for 'depth' steps.
+        3) At that point, we have the 'start' keys.
+        4) Start walking parent_map from 'start' keys, counting how many keys
+           are seen, and generating stop_keys for anything that would walk
+           outside of the parent_map.
+
+    :param parent_map: A map from {child_id: (parent_ids,)}
+    :param missing_keys: parent_ids that we know are unavailable
+    :param tip_keys: the revision_ids that we are searching
+    :param depth: How far back to walk.
+    """
+    if not parent_map:
+        # No search to send, because we haven't done any searching yet.
+        return [], [], 0
+    heads = _find_possible_heads(parent_map, tip_keys, depth)
+    s, found_heads = _run_search(parent_map, heads, set(tip_keys))
+    start_keys, exclude_keys, keys = s.get_state()
+    if found_heads:
+        # Anything in found_heads are redundant start_keys, we hit them while
+        # walking, so we can exclude them from the start list.
+        start_keys = set(start_keys).difference(found_heads)
+    return start_keys, exclude_keys, len(keys)

